#include "stdafx.h"

#include <jenga/include/smoothie/Smoothie.h>
#include <jenga/include/smoothie/LexicalAnalysis.h>

#include <Compiler.h>
#include <Procedure.h>
#include <NamespaceSupporter.h>

#include "../BasicCompiler_Common/common.h"

#ifdef _AMD64_
#include "../BasicCompiler64/opcode.h"
#else
#include "../BasicCompiler32/opcode.h"
#endif

int GetCallProcName(char *buffer,char *name){
	int i2,i3,IsStr=0;

	for(i2=0;;i2++){
		if(buffer[i2]=='\"') IsStr^=1;
		if(IsDBCSLeadByte(buffer[i2])){
			name[i2]=buffer[i2];
			i2++;
			name[i2]=buffer[i2];
			continue;
		}
		if(buffer[i2]=='['&&IsStr==0){
			i3=GetStringInBracket(name+i2,buffer+i2);
			i2+=i3-1;
			continue;
		}
		if(buffer[i2]=='('&&IsStr==0){
			name[i2]=0;
			break;
		}
		if(buffer[i2]=='='&&IsStr==0){
			name[i2]=0;
			break;
		}

		name[i2]=buffer[i2];
		if(buffer[i2]=='\0') break;
	}
	return i2;
}

int GetProc(char *name,void **ppInfo){

	//[U[`֐
	*ppInfo=(void *)GetSubHash(name);
	if(*ppInfo) return PROC_DEFAULT;

	//DLL֐
	*ppInfo=(void *)GetDeclareHash(name);
	if(*ppInfo) return PROC_DLL;

	//RpCߍ݌^
	*ppInfo=(void *)(_int64)GetFunctionFromName(name);
	if(*ppInfo) return PROC_BUILTIN;


	/////////////////////////////////////////////////////////////////
	//֐|C^A܂̓fQ[g
	/////////////////////////////////////////////////////////////////

	Type type;
	if( !GetVarType( name, type, false ) ){
		return 0;
	}

	if( type.IsProcPtr() )
	{
		// ֐|C^
		return PROC_PTR;
	}

	if( type.IsDelegate() )
	{
		// fQ[g
		return PROC_DELEGATE;
	}

	return 0;
}

void SplitObjectName(const char *name,char *ObjectName, ReferenceKind &referenceFind )
{
	referenceFind = RefNon;

	int i4;
	for(i4=lstrlen(name)-1;i4>=0;i4--){
		if(name[i4]=='.'||(name[i4]==1&&name[i4+1]==ESC_PSMEM))
			break;
	}
	if(i4==-1) ObjectName[0]=0;
	else{
		//Qƃ^Cv𔻕
		if(name[i4]=='.')
		{
			referenceFind = RefDot;
		}
		else
		{
			referenceFind = RefPointer;
		}

		if(i4==0) GetWithName(ObjectName);
		else{
			memcpy(ObjectName,name,i4);
			ObjectName[i4]=0;
		}
	}
}

bool CallProc( int kind, const void *pProc, const char *fullCallName, const char *lpszParms, const Type &baseType, Type &resultType, bool isCallOn )
{
	//GetSubHashŃG[񎦂sꂽꍇ
	if(pProc==(Procedure *)-1){
		return false;
	}

	if(kind==PROC_DEFAULT){
		/////////////////////
		// [U[`֐
		/////////////////////

		const UserProc *pUserProc = (const UserProc *)pProc;

		//IuWFNg擾
		char ObjectName[VN_SIZE];
		ReferenceKind referenceKind;
		SplitObjectName(fullCallName,ObjectName, referenceKind );


		////////////////////////
		// I[o[[h
		////////////////////////

		std::vector<const UserProc *> subs;
		GetOverloadSubHash(fullCallName,subs);
		if(subs.size()){
			//I[o[[h
			pUserProc=OverloadSolutionWithStrParam(fullCallName,subs,lpszParms,ObjectName);

			if(!pUserProc){
				return false;
			}
		}

		resultType = pUserProc->ReturnType();

		if( isCallOn ){
			if( !Opcode_CallProc(lpszParms,pUserProc,0,ObjectName ) ){
				return false;
			}
		}
	}
	else if(kind==PROC_DLL){
		/////////////////////////
		// DLL֐
		/////////////////////////
		DllProc *pDllProc = (DllProc *)pProc;

		resultType = pDllProc->ReturnType();

		if( isCallOn ){
			if( !Opcode_CallDllProc(lpszParms,pDllProc) ){
				return false;
			}
		}
	}
	else if(kind==PROC_BUILTIN){
		/////////////////////////
		// gݍ݊֐
		/////////////////////////
		int FuncId = (int)(_int64)pProc;

		if( !Opcode_CallFunc( lpszParms, FuncId, baseType, resultType, isCallOn ) ){
			return false;
		}
	}
	else if(kind==PROC_PTR){
		/////////////////
		// ֐|C^
		/////////////////

		Type type;
		GetVarType(fullCallName,type,false);

		ProcPointer *pProcPtr = compiler.GetObjectModule().meta.GetProcPointers()[type.GetIndex()];
		resultType = pProcPtr->ReturnType();

		if( isCallOn ){
			if( !Opcode_CallProcPtr(fullCallName,lpszParms,pProcPtr) ){
				return false;
			}
		}
	}
	else if( kind == PROC_DELEGATE )
	{
		// fQ[g
		char tempName[VN_SIZE];
		lstrcpy( tempName, fullCallName );
		lstrcat( tempName, ".Call" );

		void *pInfo=(void *)GetSubHash( tempName );
		if( !pInfo )
		{
			Jenga::Throw( "fQ[g̓Call\bh̎擾Ɏs" );
		}

		return CallProc( PROC_DEFAULT, pInfo, tempName, lpszParms, baseType, resultType, isCallOn );
	}
	else{
		return false;
	}

	return true;
}
bool CallPropertyMethod( const char *variable, const char *rightSide, Type &resultType){
	//vpeBp̃\bhĂяo

	//zvf擾
	char VarName[VN_SIZE],ArrayElements[VN_SIZE];
	GetArrayElement(variable,VarName,ArrayElements);

	//IuWFNg擾
	char ObjectName[VN_SIZE];
	ReferenceKind referenceKind;
	SplitObjectName(VarName,ObjectName, referenceKind );

	//I[o[[hp̊֐Xg쐬
	std::vector<const UserProc *> subs;
	GetOverloadSubHash(VarName,subs);
	if(subs.size()==0){
		return false;
	}

	//p[^𐮔
	char *Parameter;
	Parameter=(char *)HeapAlloc(hHeap,0,lstrlen(ArrayElements)+lstrlen(rightSide)+32);
	lstrcpy(Parameter,ArrayElements);
	if(rightSide){
		if(Parameter[0]&&rightSide[0]) lstrcat(Parameter,",");
		lstrcat(Parameter,rightSide);
	}

	//I[o[[h
	const UserProc *pUserProc = OverloadSolutionWithStrParam(VarName,subs,Parameter,ObjectName);

	if(pUserProc){
		//Ăяo
		Opcode_CallProc(Parameter,pUserProc,0,ObjectName);

		resultType = pUserProc->ReturnType();

		Type leftType;
		GetVarType( ObjectName, leftType, false );

		// ^p[^
		ResolveFormalGenericTypeParameter( resultType, leftType, pUserProc );
	}

	HeapDefaultFree(Parameter);

	return true;
}

bool GetReturnTypeOfPropertyMethod( const char *variable, const char *rightSide, Type &resultType ){
	//vpeBp̃\bhĂяo

	//zvf擾
	char VarName[VN_SIZE],ArrayElements[VN_SIZE];
	GetArrayElement(variable,VarName,ArrayElements);

	//IuWFNg擾
	char ObjectName[VN_SIZE];
	ReferenceKind referenceKind;
	SplitObjectName(VarName,ObjectName, referenceKind );

	//I[o[[hp̊֐Xg쐬
	std::vector<const UserProc *> subs;
	GetOverloadSubHash(VarName,subs);
	if(subs.size()==0){
		return 0;
	}

	//p[^𐮔
	char *Parameter;
	Parameter=(char *)HeapAlloc(hHeap,0,lstrlen(ArrayElements)+lstrlen(rightSide)+32);
	lstrcpy(Parameter,ArrayElements);
	if(rightSide){
		if(Parameter[0]&&rightSide[0]) lstrcat(Parameter,",");
		lstrcat(Parameter,rightSide);
	}

	//I[o[[h
	const UserProc *pUserProc = OverloadSolutionWithStrParam(VarName,subs,Parameter,ObjectName);

	if(pUserProc){
		resultType = pUserProc->ReturnType();
	}

	return 1;
}

//CfNTigetterj̖߂l擾
bool GetReturnTypeOfIndexerGetterProc( const Type &classType, Type &resultType )
{
	vector<const UserProc *> subs;
	classType.GetClass().GetMethods().Enum( CALC_ARRAY_GET, subs );
	if( subs.size() == 0 ){
		return false;
	}

	const UserProc *pUserProc = subs[0];

	resultType = pUserProc->ReturnType();


	// ^p[^
	ResolveFormalGenericTypeParameter( resultType, classType, pUserProc );

	return true;
}

void CollectProcedures( const BasicSource &source, UserProcs &userProcs, DllProcs &dllProcs )
{
	extern HANDLE hHeap;
	int i,i2,i3;
	char temporary[8192];

	// OԊǗ
	NamespaceScopes &namespaceScopes = compiler.GetNamespaceSupporter().GetLivingNamespaceScopes();
	namespaceScopes.clear();

	// ImportsꂽOԂ̊Ǘ
	NamespaceScopesCollection &importedNamespaces = compiler.GetNamespaceSupporter().GetImportedNamespaces();
	importedNamespaces.clear();

	i=-1;
	while(1){
		i++;

		if(source[i]==1&&(source[i+1]==ESC_CLASS||source[i+1]==ESC_INTERFACE)){
			/*	Class ` End Class
				Interface ` End Interface
				щz			*/
			i3=GetEndXXXCommand(source[i+1]);
			for(i+=2,i2=0;;i++,i2++){
				if(source[i]=='\0') break;
				if(source[i]==1&&source[i+1]==(char)i3){
					i++;
					break;
				}
			}
			if(source[i]=='\0') break;
			continue;
		}

		if( source[i] == 1 && source[i+1] == ESC_NAMESPACE ){
			for(i+=2,i2=0;;i2++,i++){
				if( IsCommandDelimitation( source[i] ) ){
					temporary[i2]=0;
					break;
				}
				temporary[i2]=source[i];
			}
			namespaceScopes.push_back( temporary );

			continue;
		}
		else if( source[i] == 1 && source[i+1] == ESC_ENDNAMESPACE ){
			if( namespaceScopes.size() <= 0 ){
				SetError(12, "End Namespace", i );
			}
			else{
				namespaceScopes.pop_back();
			}

			i += 2;
			continue;
		}
		else if( source[i] == 1 && source[i+1] == ESC_IMPORTS ){
			for(i+=2,i2=0;;i2++,i++){
				if( IsCommandDelimitation( source[i] ) ){
					temporary[i2]=0;
					break;
				}
				temporary[i2]=source[i];
			}
			if( !compiler.GetNamespaceSupporter().ImportsNamespace( temporary ) )
			{
				SetError(64,temporary,cp );
			}

			continue;
		}
		else if( source[i] == 1 && source[i+1] == ESC_CLEARNAMESPACEIMPORTED ){
			importedNamespaces.clear();
			continue;
		}

		if(source[i]==1&&source[i+1]==ESC_DECLARE){
			for(i+=2,i2=0;;i2++,i++){
				if(source[i]=='\n'){
					temporary[i2]=0;
					break;
				}
				temporary[i2]=source[i];
				if(source[i]=='\0') break;
			}
			dllProcs.Add(namespaceScopes,temporary,i);

			continue;
		}
		if(source[i]==1&&(source[i+1]==ESC_SUB||source[i+1]==ESC_FUNCTION||source[i+1]==ESC_MACRO)){
			char statementChar = source[i+1];

			for(i2=0;;i2++,i++){
				if(IsCommandDelimitation(source[i])){
					temporary[i2]=0;
					break;
				}
				temporary[i2]=source[i];
				if(source[i]=='\0') break;
			}
			userProcs.Add(namespaceScopes, importedNamespaces, temporary,i,false,NULL,false);

			/*	Sub ` End Sub
				Function ` End Function
				Macro ` End Macro
				щz			*/
			char endStatementChar = GetEndXXXCommand( statementChar );
			for(i2=0;;i++,i2++){
				if( source[i] == '\0' ) break;
				if( source[i] == 1 && source[i+1] == endStatementChar ){
					i++;
					break;
				}
			}
			if(source[i]=='\0') break;
			continue;
		}

		//̍s
		for(;;i++){
			if(IsCommandDelimitation(source[i])) break;
		}
		if(source[i]=='\0') break;
	}

	////////////
	// ֐
	////////////
	namespaceScopes.clear();
	importedNamespaces.clear();

	compiler.globalAreaProcName = "_System_GlobalArea_" + compiler.GetModuleName();
	sprintf(temporary,"%c%c%s()",1,ESC_SUB,compiler.globalAreaProcName.c_str());
	userProcs.Add( namespaceScopes, importedNamespaces, temporary,0,false,NULL,false);
}

bool IsNeedProcCompile(){
	compiler.GetObjectModule().meta.GetUserProcs().Iterator_Reset();
	while( compiler.GetObjectModule().meta.GetUserProcs().Iterator_HasNext() )
	{
		UserProc *pUserProc = compiler.GetObjectModule().meta.GetUserProcs().Iterator_GetNext();
		if( pUserProc->IsUsing() && pUserProc->IsCompiled() == false ){
			return true;
		}
	}
	return false;
}

void CompileBufferInProcedure( const UserProc &userProc ){
	if( userProc.IsCompiled() ) return;

	_compile_proc( &userProc );

/*
	// O𗚂
	char temporary[8192];
	temporary[0]=0;
	lstrcat( temporary, "------------------------------------------------------------------\n" );
	sprintf( temporary + lstrlen(temporary), "y %s ̃R[hz\n", userProc.GetFullName().c_str() );
	sprintf( temporary + lstrlen(temporary), "code size: %d bytes\n", userProc.GetCodeSize() );
	lstrcat( temporary, "------------------------------------------------------------------\n" );
	lstrcat( temporary, "\n" );
	Smoothie::Logger::Put( temporary );*/
}
void CompileLocal(){
	if( compiler.IsDll() )
	{
		//DLL̏ꍇ̓O[oϐ邽߂̊֐ԏ߂ɃRpC
		const UserProc *pUserProc=GetSubHash("_System_InitDllGlobalVariables");
		if(pUserProc){
			CompileBufferInProcedure( *pUserProc );
		}
		else SetError(300,NULL,cp);
	}
	else
	{
		// O[öԏ߂ɃRpC
		extern const UserProc *pSub_System_GlobalArea;
		CompileBufferInProcedure( *pSub_System_GlobalArea );
	}

	//_System_TypeBase_InitializeUserTypes͈ԍŌɃRpC
	extern const UserProc *pSubStaticMethod_System_TypeBase_InitializeUserTypes;
	pSubStaticMethod_System_TypeBase_InitializeUserTypes->CompleteCompile();

	//_System_InitStaticLocalVariables͈ԍŌɃRpC
	//ʊ֐̐ÓIϐIuWFNgׂĎWȂ΂ȂȂ
	extern const UserProc *pSub_System_InitStaticLocalVariables;
	pSub_System_InitStaticLocalVariables->CompleteCompile();

	//_System_Call_Destructor_of_GlobalObject͈ԍŌɃRpC
	extern const UserProc *pSub_System_Call_Destructor_of_GlobalObject;
	pSub_System_Call_Destructor_of_GlobalObject->CompleteCompile();

	// _System_CGarbageCollection.RegisterGlobalRoots͈ԍŌɃRpC
	extern const UserProc *pUserProc_System_CGarbageCollection_RegisterGlobalRoots;
	pUserProc_System_CGarbageCollection_RegisterGlobalRoots->CompleteCompile();

repeat:
	compiler.GetObjectModule().meta.GetUserProcs().Iterator_Reset();
	while( compiler.GetObjectModule().meta.GetUserProcs().Iterator_HasNext() )
	{
		UserProc *pUserProc = compiler.GetObjectModule().meta.GetUserProcs().Iterator_GetNext();
		CompileBufferInProcedure( *pUserProc );
	}

	if( IsNeedProcCompile() ){
		//vV[WRpCɂāAvV[WRpCKvɂȂꍇ
		goto repeat;
	}

	if( !compiler.IsStaticLibrary() )
	{
		//_System_TypeBase_InitializeUserTypes͍Ō̂قŃRpC
		pSubStaticMethod_System_TypeBase_InitializeUserTypes->KillCompileStatus();
		CompileBufferInProcedure( *pSubStaticMethod_System_TypeBase_InitializeUserTypes );

		if( IsNeedProcCompile() ){
			//vV[WRpCɂāAvV[WRpCKvɂȂꍇ

			compiler.GetObjectModule().meta.GetUserProcs().Iterator_Reset();
			while( compiler.GetObjectModule().meta.GetUserProcs().Iterator_HasNext() )
			{
				UserProc *pUserProc = compiler.GetObjectModule().meta.GetUserProcs().Iterator_GetNext();
				CompileBufferInProcedure( *pUserProc );
			}
		}

		//_System_InitStaticLocalVariables͈ԍŌɃRpC
		pSub_System_InitStaticLocalVariables->KillCompileStatus();
		CompileBufferInProcedure( *pSub_System_InitStaticLocalVariables );

		//_System_Call_Destructor_of_GlobalObject͈ԍŌɃRpC
		pSub_System_Call_Destructor_of_GlobalObject->KillCompileStatus();
		CompileBufferInProcedure( *pSub_System_Call_Destructor_of_GlobalObject );

		// _System_CGarbageCollection.RegisterGlobalRoots͈ԍŌɃRpC
		pUserProc_System_CGarbageCollection_RegisterGlobalRoots->KillCompileStatus();
		CompileBufferInProcedure( *pUserProc_System_CGarbageCollection_RegisterGlobalRoots );
	}
}
